/* glob -- glob a command's arguments, then run the command.
	   The command must be capable of processing long command lines.
   Copyright (C) 2000 Darren Salt.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:
   1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
   2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
   3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
   This product includes software developed by Darren Salt.
   4. The name of the author may not be used to endorse or promote products
   derived from this software without specific prior written permission.

   This software is provided by the author "as is" and any express or implied
   warranties including, but not limited to, the implied warranties of
   merchantability and fitness for a particular purpose are disclaimed. in no
   event shall the author be liable for any direct, indirect, incidental,
   special, exemplary, or consequential damages (including, but not limited
   to, procurement of substitute goods or services; loss of use, data, or
   profits; or business interruption) however caused and on any theory of
   liability, whether in contract, strict liability, or tort (including
   negligence or otherwise) arising in any way out of the use of this
   software, even if advised of the possibility of such damage.
 */

/* Written by Darren Salt <ds@youmustbejoking.demon.co.uk>.  */

#ifndef __riscos
#error This program only runs on RISC OS
#endif

#include "system.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <kernel.h>

#include "getopt.h"
#include "ro_wild.h"
#include "xalloc.h"
#include "error.h"

#include "ds_utils.h"

const char *program_name;

static int verbose;
static char *cmd;
static int cmd_alloc = 256;

static const char *expand_string = "{}";
static int expand_string_len = 2;

/* Display help text then exit. */
static void
help (void)
{
  printf ("\
Usage: %s [OPTIONS] COMMAND_STRING [FILESPEC]...\n\
Filename expansion and substitution in a command string.\n\
\n\
Each occurrence of STRING in the original command string is replaced by the\n\
corresponding filespec after expansion (all matching files are included).\n\
Remaining filespecs are appended after expansion, then the resulting command\n\
string is passed to the OS; redirection is passed on.\n\
\n\
The generated command string may be long; not all programs or *commands are\n\
able to cope.\n\
\n\
  -V, --verbose		verbose - print command (to stderr) before execution\n\
  -c, --c-program	this command is a C program (default)\n\
  -r, --riscos-program  this command is not a C program\n\
  -s, --string=STRING	the string to be replaced by the expanded filename(s)\n\
  -h, --help		display this help text\n\
  -v, --version		display version and licence information\n\
\n\
Bug reports to <ds@youmustbejoking.demon.co.uk>.\n\
", program_name);
  exit (0);
}


static void
version (void)
{
  printf ("glob %s\n\
\n\
Copyright (C) 2000 Darren Salt.\n\
This is free software; see the source for copying conditions. There is NO\n\
warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\
", Version);
  exit (0);
}


/* Display usage information then exit. */
static void
usage (const char *msg, const char opt)
{
  if (msg)
    fprintf (stderr, "%s: %s -- -%c\n", program_name, msg, opt);
  fprintf (stderr, "Try `%s' for more information.\n", program_name);
  exit (1);
}


static void
cmd_realloc (int i)
{
  while (strlen (cmd) + i >= cmd_alloc)
    cmd = xrealloc (cmd, cmd_alloc *= 2);
}


static void
cmd_append (int fd, int not_c, const char* dir)
{
  int namelen, dirlen;
  const char *name = fileno_to_name(fd);
  if (!name)
    error (2, 0, "out of memory");
  namelen = strlen (name);
  dirlen = strlen (dir);
  if (not_c)
    {
      cmd_realloc (namelen + dirlen + 5);
      sprintf (cmd + strlen (cmd), " { %s %s }", dir, name);
    }
  else
    {
      cmd_realloc (namelen + dirlen + 1);
      sprintf (cmd + strlen (cmd), " %s%s", dir, name);
    }
  free ((char *) name);
}


/* Return a heap block containing the first word in the string. */
static const char *
firstword (const char *str)
{
  char *p;
  int l;
  str += strspn (str, " ");
  p = strchr (str, ' ');
  if (!p)
    return xstrdup (str);
  l = (p - str);
  p = xmalloc (l + 1);
  memcpy (p, str, l);
  p[l] = 0;
  return p;
}


/* Execute a command. */
static void
exec (int not_c, char *cmd)
{
  int ret;

  cmd_realloc (6);
  memmove (cmd + 6, cmd, strlen (cmd) + 1);
  memcpy (cmd, "chain:", 6);

  /* Deal with stdin redirection */
  if (__iob[0].__file)
    cmd_append (0, not_c, "<");

  /* Deal with stdout redirection */
  if (__iob[1].__file)
    cmd_append (1, not_c, ">>");

  /* Deal with stderr redirection */
  if (!not_c && __iob[2].__file)
    {
      if (__iob[2].__file == __iob[1].__file)
	{
	  cmd_realloc (5);
	  strcat (cmd, " 2>&1");
	}
      else
	cmd_append (2, not_c, "2>>");
    }

  /* Now we can run the program... */
  if (verbose)
    fprintf (stderr, "%s\n", cmd + 6);
  _kernel_last_oserror ();
  ret = system (cmd);
  /* if we get this far, then a problem has occurred */
  if (ret < 0)
    {
      const _kernel_oserror *err = _kernel_last_oserror ();
      if (err)
	error (1, 0, "%s: %s", firstword (cmd), err->errmess);
      else
	error (1, 0, "%s: couldn't run command", firstword (cmd));
    }
  else if (ret > 0)
    error (ret, 0, "%s: exited with code %i", firstword (cmd), ret);
#undef REALLOC
}


static int
cmd_insert (int index, const char *file, int space)
{
  int filelen = strlen (file);
  cmd_realloc (filelen + space);
  memmove (cmd + index + filelen + space, cmd + index,
	   strlen (cmd) - index + 1);
  if (space)
    cmd[index] = ' ';
  memcpy (cmd + index + space, file, filelen);
  return index + filelen + space;
}


/* Wildcard expansion */
static int
expand (int index, const char *afsp)
{
  char *expand_ptr = strstr (cmd + index, expand_string);
  int new_index = expand_ptr ? (expand_ptr - cmd) : strlen (cmd);
  char *file;
  ro_wild *wild;
  int space = !expand_ptr;

  /* if the expand string is found, then new_index -> it, and space == 0
   * else new_index == commnd string length, and space == 1
   */

  if (expand_ptr)
    memmove (cmd + new_index, cmd + new_index + expand_string_len,
	     strlen (cmd) - new_index - expand_string_len + 1);

  if (!ro_wild_test (afsp))
    return cmd_insert (new_index, afsp, space);
  wild = ro_wild_init (afsp);
  while ((file = ro_wild_next (wild)) != 0)
    {
      new_index = cmd_insert (new_index, file, space);
      space = 1;
    }
  if (!ro_wild_finish (wild))
    return cmd_insert (new_index, afsp, space);
  return new_index;
}


static struct option long_options[] =
{
  {"c-program", no_argument, NULL, 'c'},
  {"riscos-program", no_argument, NULL, 'r'},
  {"string", required_argument, NULL, 's'},
  {"verbose", no_argument, NULL, 'V'},
  {"help", no_argument, NULL, 'h'},
  {"version", no_argument, NULL, 'v'},
  {NULL, 0, NULL, 0}
};


int
main (int argc, char *argv[])
{
  int c, not_c = 0;

  program_name = argv[0];
  if (argc == 1)
    help ();

  while ((c = getopt_long (argc, argv, "crs:Vhv", long_options, (int *) 0))
	 != EOF)
    switch (c)
      {
      case 'c': not_c = 0; break;
      case 'r': not_c = 1; break;
      case 's':
	if (!optarg || !optarg[0])
	  usage ("null or missing argument", 's');
	expand_string_len = strlen (expand_string = optarg);
	break;
      case 'V': verbose = 1; break;
      case 'h': help ();
      case 'v': version ();
      }

  if (optind >= argc)
    error (1, 0, "missing command string");

  cmd = xstrdup (argv[optind]);
  c = 0; /* index into cmd[] */
  while (++optind < argc)
    c = expand (c, argv[optind]);
  exec (not_c, cmd);
  return 0;
}
