/* pipe -- pipe the output of one command into the next
   Copyright (C) 1999 Darren Salt.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:
   1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
   2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
   3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
   This product includes software developed by Darren Salt.
   4. The name of the author may not be used to endorse or promote products
   derived from this software without specific prior written permission.

   This software is provided by the author "as is" and any express or implied
   warranties including, but not limited to, the implied warranties of
   merchantability and fitness for a particular purpose are disclaimed. in no
   event shall the author be liable for any direct, indirect, incidental,
   special, exemplary, or consequential damages (including, but not limited
   to, procurement of substitute goods or services; loss of use, data, or
   profits; or business interruption) however caused and on any theory of
   liability, whether in contract, strict liability, or tort (including
   negligence or otherwise) arising in any way out of the use of this
   software, even if advised of the possibility of such damage.
 */

/* Written by Darren Salt <ds@youmustbejoking.demon.co.uk>.  */

#ifndef __riscos
#error This program only runs on RISC OS
#endif

#include "system.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <kernel.h>
#include <swis.h>

#include "xalloc.h"
#include "error.h"
#include "getline.h"

#include "ds_utils.h"

const char *program_name;

static int verbose;


/* Display help text then exit. */
static void
help (void)
{
  printf ("pipe %s\n\
\n\
Copyright (C) 1999-2002 Darren Salt.\n\
This is free software; see the source for copying conditions. There is NO\n\
warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\
\n\
Usage: %s [-v] [[-c|-r] COMMAND] [[-c|-r] COMMAND]...\n\
       %s [-v] -f [FILE]\n\
Pipes output from one command into the next.\n\
\n\
  -v  verbose - print command (to stderr) before execution\n\
  -f  read input from the given file, or from standard input\n\
  -c  this command is a C program\n\
  -r  this command is not a C program, and should be run via Wimp_StartTask\n\
", Version, program_name, program_name);
  exit (0);
}


/* Display usage information then exit. */
static void
usage (const char *msg, const char *opt)
{
  printf ("\
%s: %s -- %s\n\
Try `%s' for more information.\n", program_name, msg, opt, program_name);
  exit (1);
}


/* Pipe files. */
static char *infile;
static const char *outfile;


/* Does the file exist? */
static int
isthere (const char *file)
{
  int ret;
  if (_swix (8, 3 | 1 << 31, 17, file, &ret))
    ret = 0;
  return ret;
}


/* Return a heap block containing the first word in the string. */
static const char *
firstword (const char *str)
{
  char *p;
  int l;
  str += strspn (str, " ");
  p = strchr (str, ' ');
  if (!p)
    return xstrdup (str);
  l = (p - str);
  p = xmalloc (l + 1);
  memcpy (p, str, l);
  p[l] = 0;
  return p;
}


/* Shut down, deleting any left-over pipe files. */
static void
shutdown (void)
{
  int ret;
  const char *file = infile;
  infile = 0;
  if (file && isthere (file) && remove (file))
    {
      error (0, errno, "failed to remove pipe '%s'", file);
      ret = 2;
    }
  file = outfile;
  outfile = 0;
  if (file && isthere (file) && remove (file))
    {
      error (0, errno, "failed to remove pipe '%s'", file);
      ret = 2;
    }
}


/* Execute a command. */
static int
exec (int not_c, const char *command)
{
  static char *cmd;
  char *p;
  int ret;

  /* If we have an old input file, remove it. */
  if (infile)
    {
      if (isthere (infile) && remove (infile))
	{
	  error (0, errno, "failed to remove pipe '%s'", infile);
	  ret = 2;
	}
      free (infile);
    }
  /* Make the old output file the new input file, and get a new output file.
   */
  infile = outfile ? xstrdup (outfile) : 0;
  outfile = zp_tmpnam (0);

  /* C programs are run using C-style redirection.
   * Others are run using "WimpTask Do", CLI redirection, " { > null: }".
   */
  cmd = xrealloc (cmd, strlen (command) + 128);
  if (not_c)
    sprintf (cmd, "WimpTask Do %s", command);
  else
    strcpy (cmd, command);
  p = cmd + strlen (cmd);
  if (not_c)
    {
      if (infile)
	sprintf (p, " <123> < %s }", infile);
      sprintf (p + strlen (p), " <123> > %s } { > null: }", outfile);
    }
  else
    {
      if (infile)
	sprintf (p, " < %s", infile);
      sprintf (p + strlen (p), " > %s", outfile);
    }

  /* Run the command, and if it fails, report it appropriately. */
  if (verbose)
    fprintf (stderr, "%s\n", command);
  _kernel_last_oserror ();
  ret = system (cmd);
  if (ret < 0)
    {
      const _kernel_oserror *err = _kernel_last_oserror ();
      if (err)
	error (1, 0, "%s: %s", firstword (command), err->errmess);
      else
	error (1, 0, "%s: couldn't run command", firstword (command));
    }
  else if (ret > (not_c ? 0 : 1))
    error (ret, 0, "%s: exited with code %i", firstword (command), ret);
  return ret;
}


/* Read and execute commands from an argv array. */
static int
read_from_argv (char *argv[])
{
  /* Validate arguments. */
  int i = 0, ret;
  while (argv[i])
    {
      if (argv[i][0] == '-')
	{
	  if (strcmp (argv[i], "-r") && strcmp (argv[i], "-c"))
	    usage ("unknown flag", argv[i] + 1);
	  if (!argv[++i])
	    usage ("flag requires a command string", argv[i] + 1);
	}
      i++;
    }

  /* Process arguments. */
  i = 0;
  while (argv[i])
    {
      int not_c = (argv[i][0] == '-' && argv[i++][1] == 'r');
      ret = exec (not_c, argv[i]);
      i++;
    }
  return ret;
}


/* Read commands from a file, put them in an argv array, then execute them.
 */
static int
read_from_file (const char *file)
{
  FILE *fd;
  int cmd = 0, alloc = 32;
  char **cmds = xmalloc (alloc * sizeof (char *));
  fd = file ? fopen (file, "r") : stdin;
  if (!fd)
    error (2, errno, "%s: couldn't open", file);
  do
    {
      char *p, *command = 0;
      size_t tmp;
      int len = getline (&command, &tmp, fd);
      /* EOF? */
      if (len <= 0)
	break;
      /* Lose trailing LF */
      if (command[len - 1] == '\n')
	command[len - 1] = 0;
      /* replace TABs with spaces */
      p = command;
      while ((p = strchr (p, '\t')) != 0)
	*p = ' ';
      /* strip leading spaces */
      p = command + strspn (command, " ");
      if (p != command)
	strcpy (command, p);
      /* strip trailing spaces */
      p = command + strlen (command);
      while (p > command && *--p == ' ') ;
      p[1] = 0;
      if (p < command)
	free (command);
      else
	{
	  cmds[cmd] = command;
	  if (++cmd == alloc)
	    cmds = xrealloc (cmds, (alloc += 32) * sizeof (char *));
	}
    }
  while (!feof (fd));
  fclose (fd);
  cmds[cmd] = 0;
  return read_from_argv (cmds);
}


int
main (int argc, char *argv[])
{
  int i = 1;
  int ret;
  FILE *fd;

  program_name = argv[0];

  if (argc == 1)
    help ();

  atexit (shutdown);

  if (!strcmp (argv[1], "-v"))
    {
      verbose = 1;
      i = 2;
    }

  if (!strcmp (argv[i], "-f"))
    ret = read_from_file (argv[i + 1]);
  else
    ret = read_from_argv (argv + i);

  /* Write output from last command to stdout. */
  fd = fopen (outfile, "r");
  if (!fd)
    error (2, errno, "couldn't open pipe");
  while (!feof (fd))
    {
      static char buf[256];
      int len = fread (buf, 1, 256, fd);
      if (ferror (fd))
	{
	  fclose (fd);
	  error (2, errno, "pipe read error");
	}
      fwrite (buf, 1, len, stdout);
      if (ferror (stdout))
	{
	  fclose (fd);
	  error (2, errno, "stdout write error");
	}
    }
  fclose (fd);

  return ret;
}
